/** 
Copyright (c) 2003 Helsinki Institute for Information Technology

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included 
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/



import com.hp.hpl.mesa.rdf.jena.mem.ModelMem;
import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.mesa.rdf.jena.vocabulary.*;
import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.jena.n3.*;
import com.hp.hpl.jena.daml.*;
import com.hp.hpl.jena.daml.common.*;

import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import java.io.*;
import java.util.*;


/**Annomobiili on museoesineiden automaattiseen rdf-annotointiin kehitetty ohjelma.*/
public class Annomobiili{

    private String MASAConceptFile;              
    private String xmlFile; 
    private Document document;
    private String configFileName;
    private Properties config;
    private String rangecFileName ;
    private Properties rangec;
    private String fileconfigFileName; 
    private Properties fileconf;
    protected String middlefileout; //Tm on lytn vlimuoto, kannattaisi korjata 
    protected String fileout;
    private String termfiles;
    private Vector terms;
    private Hashtable termnamespaces;
    private String museumacronym; //Kytetn URIn luomisen yhteydess
    private FileWriter writer;
    private FileReader reader;
    //    private String logfile;
    private BufferedWriter logout;
    private Enumeration interestingCardElements; 
    private Model resultModel;
    private DAMLModelImpl ontology;
    private String tokensForTokenizer;
    
 
    /** Konstruktori 
     *  - lukee konfiguraatiotiedostot,
     *  - lukee ontologian ja luo siit Jena Modelin ja
     *  - lukee XML-tiedoston ja tarkistaa sen oikeellisuuden.
     */
    public Annomobiili (String xmlFile) {
	// Luodaan Xerces DOM Parser
	DOMParser parser = new DOMParser();
	
	boolean isOntologyException = false;
	 
	// Luetaan config-tiedostot Properties-olioihin
	this.configFileName = "konfig.txt";
	this.rangecFileName = "konfig_range.txt";
	this.fileconfigFileName = "konfig_tiedosto.txt"; 
	this.middlefileout = "valimuoto.rdf"; 
	try{
	    this.writer = new FileWriter(middlefileout);
	    //this.logfile = logFile;
	    //this.logout = new BufferedWriter( new FileWriter( this.logfile ) );
	    this.config = new Properties();
	    this.config.load(new FileInputStream(configFileName));
	    this.rangec = new Properties();
	    this.rangec.load(new FileInputStream(rangecFileName));
	    this.fileconf = new Properties();
	    this.fileconf.load(new FileInputStream(fileconfigFileName));
       	
	
	    // Otetaan config-tiedostosta tiedostonimet ja muut muuttujat joita kytetn.	
	    fileout = this.fileconf.getProperty("tulostiedosto").trim();
	    museumacronym = this.fileconf.getProperty("museo").trim(); //!!!
	    termfiles = this.fileconf.getProperty("termit").trim();
	    MASAConceptFile = this.fileconf.getProperty("ontologia").trim();
	    
	    this.resultModel = new ModelMem();
	    this.ontology = new DAMLModelImpl();
	    
	    
	    // Testataan ontologian imemist DAML-modeliin.
	    // HUOM! Jos range-tsekkaus ei toimi eik ontologiatestiosuus voi syyn olla se, ett rdfs-tiedostossa
	    // ei ole oikeaa rdfs-namespacea, jonka Jena hyvksyy, eli: <!ENTITY rdfs 'http://www.w3.org/2000/01/rdf-schema#'>
	    
	    this.reader = new FileReader(MASAConceptFile);
	    System.out.println("Nyt tehdn ontologia.");
	    this.ontology = (DAMLModelImpl)(this.ontology.read(this.reader,
							       "http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-mao#"));
	    Iterator iter = this.ontology.listDAMLClasses();
	    int laskuri = 0;
	    while(iter.hasNext() && laskuri < 10){
		laskuri++;
	    }
	    if(laskuri == 0){
		isOntologyException = true;
		throw new Exception();
	    }
	}
	catch(Exception e){
	    e.printStackTrace();
	    if(isOntologyException)System.out.println("Ontologiaa ei voitu luoda. "+
						      "Tarkista, onko ontologiatiedostossa "+
						      "mritelty rdfs-nimiavaruudeksi "+
						      "http://www.w3.org/2000/01/rdf-schema# . "+
						      "(jos ei ole, korjaa se).");
	    System.exit(0);
	}

	// Parsi Dokumentti     	
	try {
	    parser.parse(xmlFile);
	    this.document = parser.getDocument();
	    //traverse (document);
	} catch (SAXException e) {
	    System.out.println ("Varmista, ett xml-tiedosto on validi. ("+e+")");
	    System.exit(0);
	} catch (IOException e) {
	    System.out.println ("Virhe tiedoston avaamisessa. ("+e+")");
	    System.exit(0);
	}
	NodeList cards = this.getCards();

	// Aseta arvot
	this.tokensForTokenizer = ":;,.()/";
    }
    

    /** Metodi hakee NodeListiin kaikki esinekortit XML:st, 
      * eli yksi kortti sislt aina yhden esineen tiedot.
      */
    protected NodeList getCards(){
	NodeList tmp = this.document.getElementsByTagName("xml-kortit");
	if(tmp.getLength()==1) {
	    return ((Element)tmp.item(0)).getChildNodes(); //palauttaa esineet, taideteokset ja muinaismuistot
	} else {
	    System.out.println ("Tarkista, ett xml-tiedostossa on vain yksi xml-korttielementti");
	    return null;
	}
	//return this.document.getElementsByTagName("esinekortti");
    }
    
    /** Metodi saa parametrina listan XML-tiedoston esinekorteista. 
      * Nm kortit kydn lpi yksi kerrallaan,
      * niist kaivetaan ulos mielenkiintoinen tieto, 
      * ja sitten haetaan termi vastaava URI
      * jos sellainen lytyy. Tulokset listn RDFModel-olioon, 
      * jonka voi kirjoittaa myhemmin RDF:ksi.
      */
    protected void createRDFModelFromXML(NodeList cardList){
	int apu = 0;
	
	//------kydn lpi kaikki kortit--------
	for (int i = 0;i < cardList.getLength();i++){
	    Node card = cardList.item(i);
	    if(!(card.getNodeName().compareTo("#text")==0)){
		System.out.println("------------------Uusi kortti (no."+(i/2+1)+")------------------");

		apu++;
		//Tehdn tietorakenne johon merkitn mitk kiinnostavat elementit lytyivt xml-kortista.
		//Myhemmin tarkistetaan checklistist, mit ei ollut, ja ksitelln sitten ne.
		interestingCardElements = config.propertyNames();
		String[][] checklist = new String[20][2];
		int littleHelper = 0;
		while(interestingCardElements.hasMoreElements()){
		    checklist[littleHelper][0] = (String)interestingCardElements.nextElement();
		    checklist[littleHelper][1] = "0";
		    littleHelper++;
		}

		//URI muodostetaan config-tiedoston museolyhenteest, joka on saatu config-tiedostosta, sek
		//XML-kortin esinenumerosta, joka koostuu pnumerosta ja alanumerosta.
		//!!!!Tehtv: Jos ei lydy numeroa tai ei ole arvo-tagia (eli tule NullPpinterException)
		//olisi hyv ksitell virhetilanne jotenkin: esim. arpomalla satunnainen uri kertomalla siit logissa.
		String thisCardUri = this.config.getProperty("nimiavaruus").trim();
		NodeList cardElements = (card).getChildNodes();
		for (int j = 0;j < cardElements.getLength();j++){  //Tss etsitn id urille.
		    Node childnode = cardElements.item(j);
		    if(childnode.getNodeName().compareTo("numero")==0){
			NodeList nroList = (((Element)childnode).getChildNodes());
			//System.out.println("thisCardUri-node:"+nroList.item(0)+"->"+nroList.item(0).getFirstChild());
			String nro = nroList.item(0).getFirstChild().getNodeValue();
			nro=nro.replace(':','_');
			thisCardUri = thisCardUri +  this.museumacronym + "_" + nro;
			j = cardElements.getLength();
		    } 
		}
		System.out.println("Kortti_id:"+thisCardUri);
		
		//------kydn lpi kaikki kortin elementit--------
		cardElements = (card).getChildNodes();
		for (int j = 0;j < cardElements.getLength();j++){  //Tss etsitn id urille.
		    Node childnode = cardElements.item(j);
		    String elementToBeProcessed = childnode.getNodeName();
		    String propUri = this.config.getProperty(elementToBeProcessed); 
		    String www_literaaliarvo = "";

		    //---turhat lapset pois
		    if(elementToBeProcessed.compareTo("#text")==0){
			//do nothing here

		    //---paikoille erityiskohtelu
		    } else if(elementToBeProcessed.compareTo("valmistuspaikka")==0 ||
			      elementToBeProcessed.compareTo("kayttopaikka")==0){
			NodeList places = childnode.getChildNodes();
			for(int k=0; k<places.getLength(); k++){
			    //System.out.println("places "+k+":"+places.item(k).getNodeName());
			    if(!(places.item(k).getNodeName().compareTo("#text")==0)) {
				NodeList value = places.item(k).getChildNodes();
				for(int l=0; l<value.getLength(); l++){
				    //System.out.println("  value "+l+":"+value.item(l).getNodeName());
				    if(!(value.item(l).getNodeName().compareTo("#text")==0)) {
					NodeList finalinfo = value.item(l).getChildNodes();					
					String info = "";
					try{
					    //for(int m=0; m<finalinfo.getLength(); m++)
					    info = info + finalinfo.item(0).getNodeValue(); 
					    //TSS LOPULLINEN ELEMENTIN ARVO
					} catch(Exception e){info = " ";}
					System.out.println("paikka:"+info);
					
					//---Asetetaan tieto literaaliarvoon
					if(www_literaaliarvo.compareTo("")==0) www_literaaliarvo = info;
					else www_literaaliarvo +=", " + info;

					//---tehdn ontologialinkitykset
					//Jos propUri==null, ei ole elementti sidottu propertyyn,
					//eli silloin ei haluta uria mihinkn, vaan ainoastaan teksti
					if (propUri!=null){

					    //Merkitn, ett on lydetty
					    //this.makeNote_elementFound(checklist, elementToBeProcessed);

					    //Haetaan linkit termitiedostosta ja listn ne rdf:n 
					    this.makeOntologyLinks(info, elementToBeProcessed, 
								   thisCardUri, propUri, checklist);
					}
				    }
				}
			    }
			}
			//Tehdn literaaliarvo
			this.makeLiteralTriple("www_"+elementToBeProcessed, www_literaaliarvo, thisCardUri);

		    //---ajoille erityiskohtelu
		    } else if(elementToBeProcessed.compareTo("valmistusaika")==0) {
			NodeList times = childnode.getChildNodes();
			for(int k=0; k<times.getLength(); k++){
			    //System.out.println("places "+k+":"+places.item(k).getNodeName());
			    if(!(times.item(k).getNodeName().compareTo("#text")==0)) {
				NodeList value = times.item(k).getChildNodes();
				if(value.getLength()>1 && !(www_literaaliarvo.compareTo("")==0)) www_literaaliarvo += ", "; 
				for(int l=0; l<value.getLength(); l++){
				    //System.out.println("  value "+l+":"+value.item(l).getNodeName());
				    if(!(value.item(l).getNodeName().compareTo("#text")==0)) {
					NodeList finalinfo = value.item(l).getChildNodes();
					String timename = value.item(l).getNodeName();
					String info = "";
					try{
					    //for(int m=0; m<finalinfo.getLength(); m++)
					    info = info + finalinfo.item(0).getNodeValue(); 
					    //TSS LOPULLINEN ELEMENTIN ARVO
					} catch(Exception e){info = " ";}
					System.out.println("aika:"+timename+"="+info);
					
					//---tehdn ripustus !!Huom. Ripustus erilainen kuin muissa tapauksissa
					this.makeLiteralTriple(elementToBeProcessed+"_"+timename, info, thisCardUri);

					//---Asetetaan tieto literaaliarvoon
					if(timename.compareTo("alku")==0) www_literaaliarvo += info+"-";
					else if(timename.compareTo("loppu")==0) www_literaaliarvo +="-"+info;
				    }
				}
			    }
			}
			if(www_literaaliarvo.endsWith(", ")) 
			    www_literaaliarvo = www_literaaliarvo.substring(0, www_literaaliarvo.length()-2);
			www_literaaliarvo = www_literaaliarvo.replaceAll("--","-");
			this.makeLiteralTriple("www_"+elementToBeProcessed, www_literaaliarvo, thisCardUri);

		    //---kuville erityiskohtelu
		    } else if(elementToBeProcessed.compareTo("kuvatiedostonimi")==0) {
			NodeList value = childnode.getChildNodes();
			for(int k=0; k<value.getLength(); k++){
			    //System.out.println("places "+k+":"+places.item(k).getNodeName());
			    if(!(value.item(k).getNodeName().compareTo("#text")==0)) {
				NodeList finalinfo = value.item(k).getChildNodes();	
				String info = "";
				try{
				    //for(int m=0; m<finalinfo.getLength(); m++)
				    info = info + finalinfo.item(0).getNodeValue(); 
				    //TSS LOPULLINEN ELEMENTIN ARVO
				} catch(Exception e){info = " ";}
				System.out.println(elementToBeProcessed+":"+info);
				//Tehdn literaaliarvo
				this.makeLiteralTriple("www_"+elementToBeProcessed, info, thisCardUri);
			    }
			}

		    //---loput ksiteltvt elementit
		    } else {
			NodeList value = childnode.getChildNodes();
			for(int k=0; k<value.getLength(); k++){
			    //System.out.println("places "+k+":"+places.item(k).getNodeName());
			    if(!(value.item(k).getNodeName().compareTo("#text")==0)) {
				NodeList finalinfo = value.item(k).getChildNodes();	
				String info = "";
				try{
				    //for(int m=0; m<finalinfo.getLength(); m++)
				    info = info + finalinfo.item(0).getNodeValue(); 
				    //TSS LOPULLINEN ELEMENTIN ARVO
				} catch(Exception e){info = " ";}
				System.out.println(elementToBeProcessed+":"+info);

				//---Asetetaan tieto literaaliarvoon
				if(www_literaaliarvo.compareTo("")==0) www_literaaliarvo = info;
				else www_literaaliarvo +=", " + info;
				
				//---tehdn ontologialinkitykset
				//Jos propUri==null, ei ole elementti sidottu propertyyn,
				//eli silloin ei haluta uria mihinkn, vaan ainoastaan teksti
				if (propUri!=null){
				    
				    //Merkitn, ett on lydetty
				    //this.makeNote_elementFound(checklist, elementToBeProcessed);
				    
				    //Haetaan linkit termitiedostosta ja listn ne rdf:n 
				    this.makeOntologyLinks(info, elementToBeProcessed, 
							   thisCardUri, propUri, checklist);
				}
			    }
			}
			//Tehdn literaaliarvo
			this.makeLiteralTriple("www_"+elementToBeProcessed, www_literaaliarvo, thisCardUri);
		    }
		}
		//---tehdn lopputarkistukset ja jutut kortille:
		//Varmistetaan, ett kaikki ontologialinkityksille olennaiset kentt lytyivt...
		//...eli tss vliss tarkastetaan viel, mitk elementit ovat puuttuneet kokonaan xml-kortista 
		//tai mille elementeille ei lytynyt yhtn ripustusta (ja joille siis halutaan ripustukset) ja
		//listn nist tyhj-arvot.
		int littleHelper3 = 0;
		String addMe = "";
		try{
		    while (littleHelper3 < 20){
			if (checklist[littleHelper3][0]!=null){
			    if (checklist[littleHelper3][1].equals("0")){
				addMe = checklist[littleHelper3][0];
				String propUri2 = this.config.getProperty(addMe); 
				String unknownuri;
				String rangeuri = this.getSuperRange(addMe);
				if(rangeuri!=null){
				    //System.out.println("Yuhuu. Listn koska lytyi superrange.");
				    Statement triple;
				    triple = new StatementImpl(new ResourceImpl(thisCardUri),
							       new PropertyImpl(propUri2),
							       new ResourceImpl((String)(rangeuri)));
				    this.resultModel.add(triple);
				}
			    }
			}
			littleHelper3++;
		    }
		}
		catch(Exception e){
		    e.printStackTrace();
		}
		//Kirjoitetaan kortti tiedostoon
		this.writeAllToFile(); //Kortti kirjoitetaan aina yksi kerrallaan, ettei Jenalta lopu muisti.
	    }
	}
	    
	try{
	    //logout.close();
	    this.fileconf.store(new FileOutputStream(this.fileconfigFileName),null);
	}
	catch(Exception e){
	    e.printStackTrace();
	}
    }
    
    private void writeAllToFile(){
	try{
	    this.resultModel.write(this.writer, "RDF/XML", null);
            this.resultModel = null;
	    this.resultModel = new ModelMem();
	}
	catch(Exception e){
	    e.printStackTrace();
	}
    }


    /** Metodi etsii termitiedostoista kandidaatteja ontologiaripustuksille.
      * Lydetyist kandidaateista varmistetaan, ett ne:
      * 1)  tyttvt konf_range-asetustiedostossa mritellyt ripustusrajoitukset ja
      * 2)  ettei ripustuksiin list samaan haaraan kuuluvia ontologiaripustuksia,
      *     esimerkiksi ettei list ripustusta kenkiin, jos mukana on mys ripustus
      *     korkokenkiin.
      * Ne urit, jotka tyttvt em. ehdot, listn ontologisiksi ripustuksiksi esineelle. 
      */
    private void makeOntologyLinks(String info, String elementToBeProcessed, 
				   String thisCardUri, String propUri, String[][] checklist) {
	StringTokenizer tok = new StringTokenizer(info, tokensForTokenizer);
	Vector urisBoundToElement = new Vector();
	while(tok.hasMoreTokens()){
	    String term = tok.nextToken().trim();
	    String newUri = this.getURI(term,this.termfiles);
	    System.out.println("  Mahdolliset ripustukset termille \""+term+"\" :"+newUri+".");
	    if(newUri!=null){
		int urilaskuri = 0;
		StringTokenizer st = new StringTokenizer(newUri,"&");
		while(st.hasMoreTokens()){
		    newUri = st.nextToken();
		    //System.out.println("checkRange"+this.checkRange(newUri, elementToBeProcessed));
		    if(this.checkRange(newUri,elementToBeProcessed)){
			//System.out.println("checkUri"+this.checkUri(urisBoundToElement, newUri));
			if(this.checkUri(urisBoundToElement,newUri)) {
			    //Merkitn, ett on lydetty
			    this.makeNote_elementFound(checklist, elementToBeProcessed);				    
			    //Tarkistetaan, onko osumia useampi kuin yksi eli onko sana moniselitteinen
			    urilaskuri++;
			    if(urilaskuri>1){
				try {
				    Statement triple = new StatementImpl(new ResourceImpl(thisCardUri),
									 new PropertyImpl(this.config.
											  getProperty("nimiavaruus").trim()+
											  "comment"),
									 new LiteralImpl(("Termille "+term+
				    							  " lytyi monta osumaa."+
				    							  " Ripustus on syyt"+
				    							  " tarkistaa.")));
				    this.resultModel.add(triple);
				    System.out.println("  Moniselitteinen termi: "+term +": "+triple);

				} catch (Exception e) {
				    System.out.println("  HUOMIO: Termille "+term+" lytyi monta osumaa."+
						       "Ripustus on syyt tarkistaa. "+e);
				}
			    }
			}
		    }
		}
	    } else { //Eli yhtn uria ei lytynyt
		try {
		    Statement triple = new StatementImpl(new ResourceImpl(thisCardUri),
							 new PropertyImpl(this.config.getProperty("nimiavaruus").trim()+
									  "comment"),
							 new LiteralImpl(("Termille "+term+
								 	  " ei lytynyt yhtn ontologista"+
									  " ripustusta. Kannattaa tytt"+
							 		  " termikortti termiontologiaan.")));
		    this.resultModel.add(triple);
		    System.out.println("  Termikortti puuttuu: "+term);
		} catch(Exception e) {
		    System.out.println("  Huomio: Termille "+term+" ei lytynyt yhtn ontologista ripustusta."+
				       "Kannattaa tytt termikortti termiontologiaan.");
		}
	    }
	}
	try{
	    if(urisBoundToElement.size()>0){
		ListIterator iitter = urisBoundToElement.listIterator(); 
		Statement triple;
		while(iitter.hasNext()){
		    String goodUri = (String)iitter.next();
		    //System.out.println(goodUri);
		    triple = new StatementImpl(new ResourceImpl(thisCardUri),
					       new PropertyImpl(propUri),
					       new ResourceImpl(goodUri));
		    this.resultModel.add(triple);
		    System.out.println("  Listty:"+thisCardUri+"->"+
				       propUri+"->"+goodUri);
		    //this.writeAllToFile();
		}
	    } else { //Ei ollut yhtn linkityst
		//thn vliin mahdolliset tuntemattomien lisykset, jos sellaisia halutaan
	    }
	} catch(Exception e){
	    e.printStackTrace();
	}
    }

    private void makeLiteralTriple(String elementName, String www_literaaliarvo, String thisCardUri) {
	try{
	    Statement wwwtriple;
	    String www_propUri = this.config.getProperty("nimiavaruus").trim() + elementName;
	    wwwtriple = new StatementImpl(new ResourceImpl(thisCardUri),
					  new PropertyImpl(www_propUri),
					  new LiteralImpl(www_literaaliarvo));
	    this.resultModel.add(wwwtriple);
	    //this.writeAllToFile();
	}
	catch(Exception e){
	    e.printStackTrace();
	}
    }

    private void makeNote_elementFound(String[][] checklist, String elementToBeProcessed) {
	int littleHelper2 = 0;
	while(littleHelper2 < 20){
	    if(checklist[littleHelper2][0]!=null){
		if(checklist[littleHelper2][0].equals(elementToBeProcessed)) {
		    //System.out.println("Kyll tst xml-kortista lytyi tm " + elementToBeProcessed);
		    checklist[littleHelper2][1] = "1";
		}
	    }
	    littleHelper2++;
	}
    }

    
    /** Metodi poistaa valmiista RDF-tiedostosta ylimriset tagit, jotka syntyvt 
      * koska jokainen kortti pit tallentaa omana "RDF-tiedostonaan" ettei Jenan ModelMem
      * veisi liikaa muistia ja ohjeolma kaatuisi. Tt tehtess j tiedostoon joka vliin
      * vri RDF-alku- ja -lopputageja, jotka poistetaan lopuksi tll metodilla 
      * tiedostosta ja kirjoitetaan toimiva tavara uuteen tiedostoon
      */
    protected void cleanFileFromCrap(String tiedosto, String tiedosto2){
       
	    //    Runtime annomobiili = Runtime.getRuntime();
	    //Process uriprocess = annomobiili.exec("/home/group/seco/public_html/"+
	    //				  "intra/Museo/Java/Annomobiili/poistaRDFtagit.pl < "+ 
	    //				  middlefileout + " > " + resultFile);
	Vector tulos = new Vector();
	try{
	    //System.out.println("Yuhuu.");
	    String line = null;
	    String rivi;
	    BufferedReader br = new BufferedReader(new FileReader(tiedosto));
	    BufferedWriter out = new BufferedWriter( new FileWriter( tiedosto2 ) );
	    boolean alkuloppu = false;
	    while ((line = br.readLine()) != null) {
		//System.out.println(line);
		//	tulos.add(line);    
		rivi = line;
		if(rivi.indexOf("<rdf:Description")!=-1){
		    rivi = "<NS0:Esinekortti" + rivi.substring(rivi.indexOf("n")+1,rivi.length());
		    //System.out.println(rivi + " ON SIIS SE KORJATTU");
		}
		if(rivi.indexOf("</rdf:Description")!=-1){
		    rivi = "</NS0:Esinekortti" + rivi.substring(rivi.indexOf("n")+1,rivi.length());
		    //System.out.println(rivi + " ON SIIS SE KORJATTU");
		}
		if (!alkuloppu && rivi.indexOf("Esinekortti")==-1){ 
		    
		    out.write(rivi); //Tm jotta alun rdf-mrittelyt eivt lhde pois.
		    out.newLine();
		    //System.out.println("Printattiin tiedostoon " + rivi);
		}
		if (!alkuloppu && rivi.indexOf("Esinekortti")!=-1){
		    alkuloppu=true;
		}
		if(alkuloppu){
		    if(rivi.indexOf("NS0:")!=-1 || rivi.indexOf("Esinekortti")!=-1) {
			out.write(rivi);
			out.newLine();
		    }
		}
	    }
	    br.close();                
	    br = null;
	    out.write("</rdf:RDF>");
	    out.flush();
	    out.close();
	}
	catch (java.io.FileNotFoundException ex) {
	    System.out.println("File does not exist. ");
	} 
	catch (java.io.IOException ex) {
	    ex.printStackTrace();
	} 
    }
    
    
    /** Metodille annetaan Vector edellisist ureista, 
      * ja uusi mahdollinen uri. Tsekataan onko uusi uri vanhojen aliluokka, 
      * jos on, korvataan yllokka alaluokalla. Jos taas uusi uri on vanhojen 
      * luokkien ylluokka, ei tehd mitn. Jos uusi uri ei liity vanhoihin mitenkn, 
      * se listn vectoriin. Palauttaa true jos listaan tulee lis tavaraa, false jos ei.
      */
    private boolean checkUri(Vector uris, String newUri){
	ListIterator iter = uris.listIterator();
	ListIterator secondRound = uris.listIterator();
	Iterator subs = null;
	Iterator subs2 = null;
	int helper = 0;
	while (iter.hasNext()){
	    String uriInTurn = (String)iter.next();
	    DAMLClass cls = new DAMLClassImpl(uriInTurn, 
					      (DAMLModelImpl)(this.ontology),
					      com.hp.hpl.jena.vocabulary.DAML_OIL.getInstance());
	    /*
	     subs = cls.getSubClasses();
	     while (subs.hasNext()){
		 if((((Resource)subs.next()).getURI()).compareTo(newUri)==0){
		     System.out.println("Jit kiinni!");
		     uris.set(helper,newUri);
		     return true;
		 }
		 else {
		     System.out.println("Dang.");
		     helper++;
		 }
		}
	    */
	    
	    if(cls.hasSubClass(new ResourceImpl(newUri))){
		//System.out.println("Jit kiinni!");
		uris.set(helper,newUri);
		return true;
	    }
	    else {
		//System.out.println("Dang.");
		helper++;
	    }
	}
	//System.out.println("No ei ollut. Ents toisinpin?");
	DAMLClass cls2 = new DAMLClassImpl(newUri, 
					   (DAMLModelImpl)(this.ontology),
					   com.hp.hpl.jena.vocabulary.DAML_OIL.getInstance());
	//secondRound2 = cls2.getSubClasses(); 
	while (secondRound.hasNext()){
	    String uriInTurn = (String)secondRound.next();
	    //System.out.println(uriInTurn + " on siis se johon verrataan uri-vectorin kamaa.");
	    if(cls2.hasSubClass(new ResourceImpl(uriInTurn))){
		//System.out.println("Jit kiinni toisella kierroksella! Mitn ei muuteta.");
		return false;
	    }
	}
	//System.out.println("No ei ollut niinkn. Sitten voidaan list" + newUri);
	uris.add(newUri);
	return true;	    	
    }

    /** Tm metodi palauttaa tietyn haluttavan elementin range-luokan. 
      * Kytetn jotta saadaan aihepiirin yleisin luokka laitettua
      * arvoksi silloin, kun muita ureja ei lydy.
      */
    private String getSuperRange(String elementname){
	String uri = this.rangec.getProperty(elementname);
	if (uri == null || uri.indexOf("(ontology")!=-1) return null;
	else return uri;
    }


    /** Metodilla tarkistetaan ett uri sopii rangeConfig-tiedostossa olevaan
      * elementtiin sidottuun rangeen.
      * a) Tll hetkell arvon pit olla rangeksi mritellyn urin ALALUOKKA, 
      *    eli itse range-luokka ei ky. Jos rangea ei olla sidottu rangeconf-tiedostossa, 
      *    ei rajoitteita ole, ja palautetaan true, kuten mys silloin jos uri osuu rangeen.
      *    Ainoastaan olemassaolevan mritelmn vastainen range hyltn. 
      * b) Jos rangen perss lukee(ontology), tarkistetaan ainoastaan, ett uri on 
      *    samasta ontologiasta kuin rangessa on mritelty. Koska esimerkiksi paikat 
      *    ovat omassa ontologiassaan, voidaan kayttopaikka sitoa paikat-ontologiaan, 
      *    koska halutaan ett sen arvoiksi tulee ainoastaan paikkaontologian instansseja. 
      */
    private boolean checkRange(String uriToBeChecked, String elementname){
	String rangeuri = this.rangec.getProperty(elementname);
	if (rangeuri!=null){
	    //System.out.println("rangeuri:"+rangeuri);
	    if(rangeuri.endsWith("(ontology)")){
		StringTokenizer tok = new StringTokenizer(rangeuri, ",");
		int risuaitapaikka = uriToBeChecked.indexOf("#");
		uriToBeChecked = uriToBeChecked.substring(0,risuaitapaikka);
		System.out.println("  Tarkistetaan uri:"+uriToBeChecked);
		System.out.println("  Mahdollisia luokkarajoja on "+ tok.countTokens()+"kpl.");
		while(tok.hasMoreTokens()) {
		    //System.out.println("");
		    rangeuri = tok.nextToken().trim();
		    //System.out.println("rangeuri1:"+rangeuri);
		    //System.out.println(rangeuri);
		    //rangeuri = this.cutBraces(rangeuri);
		    //System.out.println(risuaitapaikka);
		    System.out.println("  Vertailussa luokkaja:"+rangeuri+"=="+uriToBeChecked);
		    if(rangeuri.compareTo(uriToBeChecked+"(ontology)")==0){ 
			//System.out.println("Range ok.");
			return true;
		    }
		}
		return false;
	    } else {		
		DAMLClass cls = new DAMLClassImpl(rangeuri, 
						  (DAMLModelImpl)(this.ontology),
						  com.hp.hpl.jena.vocabulary.DAML_OIL.getInstance());
		if(cls.hasSubClass(new ResourceImpl(uriToBeChecked))){
		    //System.out.println("Sopii rangeen.");
		    return true;
		} else {
		    System.out.println("  Ei sittenkn listty, koska luokka \""+
				       uriToBeChecked +"\" ei ole luokan \""+
				       rangeuri+"\" alaluokka."); 

		    return false;
		}
	    }
	}
	return true; //Jos mitn rajoitusta ei ole annettu, oletetaan kaiken kyvn.
    }

    
    /** Metodi tekee saman kuin edellinen metodi, mutta ilman perli, eli hakee termiin 
      * sopivia ureja termitiedostosta. Ratkaisu on perin tehoton, optimoikoon ken tahtoo.
      * Metodi on kuitenkin tarkoitus ajaa aineistolle vain kerran, joten hitaus ei sinns haittaa.
      */
    private String getURI(String term, String termfiles){
	//System.out.println("get Urissa");
	StringTokenizer st = new StringTokenizer(termfiles,",");
	String uris = "";
	try {
	    while(st.hasMoreTokens()){
		String file = st.nextToken();
		file = file.trim();
		if((term==null) || (term.length() < 1)) return null;
		term = term.toLowerCase();
		this.termnamespaces = new Hashtable();
		Vector namespaces = new Vector();
		String line = null;
		BufferedReader br = new BufferedReader(new FileReader(file));
		String firstletter = term.substring(0,1);
		boolean termsBegun = false;
		boolean hasFoundTerm = false;
		firstletter = firstletter.toUpperCase();
		while ((line = br.readLine()) != null) {
		    if(line.indexOf("ENTITY")!=-1 && !termsBegun){ //Napataan namespacet talteen.
			namespaces.add(line);
			//System.out.println(line);
		    }
		    if(namespaces.size() > 0 && line.indexOf("ENTITY")==-1 && !termsBegun){
			ListIterator iter = namespaces.listIterator(); //Laitetaan namespacet hajautustauluun
			while(iter.hasNext()){
			    String key,value;
			    String rivi = ((String)iter.next()).trim();
			    //System.out.println(rivi.lastIndexOf(" "));
			    key = rivi.substring(9,rivi.lastIndexOf(" "));
			    //System.out.println(key);
			    value = rivi.substring(rivi.lastIndexOf(" ")+2,rivi.length()-2);
			    //System.out.println(value);
			    termnamespaces.put(key,value);
			}
			//System.out.println("Hash tehty");
			termsBegun = true;
		    }
		    if(termsBegun){
			String rivi = line;
			String vertailtava = rivi.toLowerCase();
			if(!hasFoundTerm && vertailtava.indexOf("\"" + term + "\"")!=-1)
			    hasFoundTerm = true;
			if(!hasFoundTerm && vertailtava.indexOf(">" + term + "<")!=-1)
			    hasFoundTerm = true;
			if(hasFoundTerm){
			    if(rivi.indexOf("concept")!=-1){
				String key, uri_id;
				//System.out.println(rivi);
				key = rivi.substring(rivi.indexOf("&")+1,rivi.indexOf(";"));
				//System.out.println(key);
				//System.out.println(rivi.indexOf("\""));
				uri_id = rivi.substring(rivi.indexOf(";")+1,rivi.lastIndexOf("\""));
				//System.out.println(uri_id);
				//Ei laiteta tuplaa mukaan
				if(uris.indexOf(termnamespaces.get(key) + uri_id + "&")==-1) {
				    uris = uris.concat(termnamespaces.get(key) + uri_id + "&"); //& on erotinmerkki ureille.
				    //System.out.println("listty uris:"+uris);
				} else ; //System.out.println("ei listty uris:"+uri_id);
				//System.out.println("uris:"+uris);
				hasFoundTerm = false;
			    }
			    if(rivi.indexOf("rdf:about")!=-1) hasFoundTerm = false;
			}
		    }
		}
		br.close();                
		br = null;
	    }

	    //System.out.println("get Urin lopussa");
	    if(uris.compareTo("")==0 ^ uris.compareTo("fail")==0) {
		return null;
	    } else {
		//System.out.println("Tll poistetaan uri-ptkst viimeinen turha erotinmerkki &.");
		//System.out.println(uris);
		uris = uris.substring(0,(uris.length()-1));
		//System.out.println(uris);
		return uris;
	    }
	} catch(Exception e){
	    e.printStackTrace();
	    return null;
	}
    }

    
    public static void main(String[] totuus){
	try{
	    System.out.println("");
	    System.out.println("");
	    System.out.println("Pois alta! Tlt tulee Annomobiili (versio 1.0)!");    
	    System.out.println("");
	    System.out.println("                    -----------");              
	    System.out.println("                    |         |");
	    System.out.println("                    |         |");
	    System.out.println("                   &|         |");
	    System.out.println("                   O-------------------------|");
	    System.out.println("      |-------|     |         |              |");
	    System.out.println("      |       |----/| --      |              |");
	    System.out.println("      | _____ |     |         |     _____    |");
	    System.out.println("       /     \\\\     |         |    /     \\ _/");
	    System.out.println("       | /\\  | \\-------------------| /\\  |/  ");
	    System.out.println("       | \\/  |                     | \\/  |   ");
	    System.out.println("       \\_____/                     \\_____/");
	    System.out.println("");
	    
	    
            if(totuus==null || totuus.length==0 || totuus[0]==null) {
                System.out.println("");
                System.out.println("Tarvitsen kutsun yhteydess parametrina: ");
                System.out.println("  polun esinetietoja sisltvn XML-tiedostoon.");
                System.out.println("");
            } else {
		Annomobiili hurja = new Annomobiili(totuus[0]); //XML-filename		
		hurja.createRDFModelFromXML(hurja.getCards());
		hurja.cleanFileFromCrap(hurja.middlefileout,hurja.fileout);
		//System.out.println(hurja.getURIinJava("senni",hurja.termfiles));
		System.out.println("");
		System.out.println("Olen valmis. Kiitos ajastasi.");
	    }
	}
	catch(Throwable e){
	    e.printStackTrace();
	}
    }
    
}
